/*
 * timeshift.c:
 * Time-shifting shared object.
 *
 * Copyright (c) 2003 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: timeshift.c,v 1.3 2003/05/21 11:54:46 chris Exp $";

/* needed for tue RTLD_ constants for some reason */
#define _GNU_SOURCE

#include <sys/types.h>

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/time.h>

static int is_shifted;
static time_t shift_start, shift_end;

/* setup_shift
 * If the environment variable TIMESHIFT is set and consists of two epoch
 * times separated by a :, enable time shifting and set up the window into
 * which real time is mapped. */
static inline void setup_shift(void) {
    char *s;
    if ((s = getenv("TIMESHIFT"))) {
        long int start, end;
        if (sscanf(s, "%ld:%ld", &start, &end) == 2 && end > start) {
            is_shifted = 1;
            shift_start = (time_t)start;
            shift_end   = (time_t)end;
        }
    }
}

/* shift TIME
 * If a time shift is enabled, modify *TIME to lie in the designated window.
 * Otherwise, do nothing. */
static inline void shift(time_t *t) {
    if (!is_shifted)
        return;
    *t -= shift_start;
    *t %= (shift_end - shift_start);
    *t += shift_start;
}

/*
 * Functions which substitute for libc functions. We care about time(3) and
 * gettimeofday(2). Strictly the latter is a syscall, but we override the
 * libc wrapper with our own function. Later glibc versions might break this.
 */

time_t time(time_t *T) {
    static time_t (*real_time)(time_t *);
    time_t x;
    if (!real_time) {
        real_time = dlsym(RTLD_NEXT, "time");
        setup_shift();
    }
    if (real_time(&x) == -1)
        return -1;
    shift(&x);
    if (T)
        *T = x;
    return x;
}

int gettimeofday(struct timeval *tv, struct timezone *tz) {
    static int (*real_gettimeofday)(struct timeval *tv, struct timezone *tz);
    struct timeval tvx;
    if (!real_gettimeofday) {
        real_gettimeofday = dlsym(RTLD_NEXT, "gettimeofday");
        setup_shift();
    }
    if (real_gettimeofday(&tvx, tz) == -1)
        return -1;
    shift(&tvx.tv_sec);
    *tv = tvx;
    return 0;
}


